'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var codegen = require('@smikhalevski/codegen');
var changeCaseAll = require('change-case-all');
var jtdDialect = require('@jtdc/jtd-dialect');

/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */

var __assign = function() {
    __assign = Object.assign || function __assign(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p)) t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};

function createMap() {
    return Object.create(null);
}
function die(message) {
    throw new Error(message);
}

/**
 * Converts JTD and its dependencies to a map of nodes where key is `ref` and value is a parsed node.
 *
 * @param ref The ref of the root JTD.
 * @param jtdRoot The JTD to parse.
 *
 * @returns The map from ref to a parsed node.
 */
function parseJtdRoot(ref, jtdRoot) {
    var nodes = jtdRoot.definitions ? parseJtdDefinitions(jtdRoot.definitions) : createMap();
    nodes[ref] = parseJtd(jtdRoot);
    return nodes;
}
/**
 * Converts JTD dependencies to a map of nodes where key is `ref` and value is a parsed node.
 *
 * @param jtdDefinitions The dictionary of ref-JTD pairs.
 *
 * @returns The map from ref to a parsed node.
 */
function parseJtdDefinitions(jtdDefinitions) {
    var nodes = createMap();
    for (var _i = 0, _a = Object.entries(jtdDefinitions); _i < _a.length; _i++) {
        var _b = _a[_i], ref = _b[0], jtd = _b[1];
        nodes[ref] = parseJtd(jtd);
    }
    return nodes;
}
/**
 * Converts JTD to a corresponding node.
 *
 * @param jtd The JTD to parse.
 *
 * @returns A parsed node.
 *
 * @see https://tools.ietf.org/html/rfc8927 RFC8927
 * @see https://jsontypedef.com/docs/jtd-in-5-minutes JTD in 5 minutes
 */
function parseJtd(jtd) {
    var jtdNullable = jtd.nullable, jtdType = jtd.type, jtdRef = jtd.ref, jtdEnum = jtd.enum, jtdElements = jtd.elements, jtdValues = jtd.values, jtdProperties = jtd.properties, jtdOptionalProperties = jtd.optionalProperties, jtdDiscriminator = jtd.discriminator, jtdMapping = jtd.mapping;
    if (jtdNullable) {
        var node = {
            nodeType: 2 /* NULLABLE */,
            valueNode: parseJtd(__assign(__assign({}, jtd), { nullable: undefined })),
            parentNode: null,
            jtd: jtd,
        };
        node.valueNode.parentNode = node;
        return node;
    }
    if (jtdType) {
        return {
            nodeType: 3 /* TYPE */,
            type: jtdType,
            parentNode: null,
            jtd: jtd,
        };
    }
    if (jtdRef) {
        return {
            nodeType: 1 /* REF */,
            ref: jtdRef,
            parentNode: null,
            jtd: jtd,
        };
    }
    if (jtdEnum) {
        return {
            nodeType: 4 /* ENUM */,
            values: jtdEnum,
            parentNode: null,
            jtd: jtd,
        };
    }
    if (jtdElements) {
        var node = {
            nodeType: 5 /* ELEMENTS */,
            elementNode: parseJtd(jtdElements),
            parentNode: null,
            jtd: jtd,
        };
        node.elementNode.parentNode = node;
        return node;
    }
    if (jtdValues) {
        var node = {
            nodeType: 6 /* VALUES */,
            valueNode: parseJtd(jtdValues),
            parentNode: null,
            jtd: jtd,
        };
        node.valueNode.parentNode = node;
        return node;
    }
    if (jtdProperties || jtdOptionalProperties) {
        var objectNode = {
            nodeType: 7 /* OBJECT */,
            properties: createMap(),
            optionalProperties: createMap(),
            parentNode: null,
            jtd: jtd,
        };
        if (jtdProperties) {
            for (var _i = 0, _a = Object.entries(jtdProperties); _i < _a.length; _i++) {
                var _b = _a[_i], propKey = _b[0], propJtd = _b[1];
                var propNode = objectNode.properties[propKey] = parseJtd(propJtd);
                propNode.parentNode = objectNode;
            }
        }
        if (jtdOptionalProperties) {
            for (var _c = 0, _d = Object.entries(jtdOptionalProperties); _c < _d.length; _c++) {
                var _e = _d[_c], propKey = _e[0], propJtd = _e[1];
                if (propKey in objectNode.properties) {
                    die('Duplicated property: ' + propKey);
                }
                var propNode = objectNode.optionalProperties[propKey] = parseJtd(propJtd);
                propNode.parentNode = objectNode;
            }
        }
        return objectNode;
    }
    if (jtdDiscriminator || jtdMapping) {
        if (!jtdDiscriminator || !jtdMapping) {
            die('Malformed discriminated union');
        }
        var unionNode = {
            nodeType: 8 /* UNION */,
            discriminator: jtdDiscriminator,
            mapping: createMap(),
            parentNode: null,
            jtd: jtd,
        };
        for (var _f = 0, _g = Object.entries(jtdMapping); _f < _g.length; _f++) {
            var _h = _g[_f], mappingKey = _h[0], mappingJtd = _h[1];
            var objectNode = parseJtd(mappingJtd);
            if (objectNode.nodeType !== 7 /* OBJECT */) {
                die('Mappings must be object definitions: ' + mappingKey);
            }
            unionNode.mapping[mappingKey] = objectNode;
            objectNode.parentNode = unionNode;
        }
        return unionNode;
    }
    return {
        nodeType: 0 /* ANY */,
        parentNode: null,
        jtd: jtd,
    };
}

/**
 * Invokes callbacks from `visitor` for each node tree under `node`.
 *
 * @param node The root of the JTD node tree.
 * @param visitor Callbacks that must be invoked for nodes under `node`.
 */
function visitJtdNode(node, visitor) {
    var _a, _b, _c, _d, _e, _f, _g, _h, _j;
    switch (node.nodeType) {
        case 0 /* ANY */:
            (_a = visitor.any) === null || _a === void 0 ? void 0 : _a.call(visitor, node);
            break;
        case 1 /* REF */:
            (_b = visitor.ref) === null || _b === void 0 ? void 0 : _b.call(visitor, node);
            break;
        case 2 /* NULLABLE */:
            (_c = visitor.nullable) === null || _c === void 0 ? void 0 : _c.call(visitor, node, function () { return visitJtdNode(node.valueNode, visitor); });
            break;
        case 3 /* TYPE */:
            (_d = visitor.type) === null || _d === void 0 ? void 0 : _d.call(visitor, node);
            break;
        case 4 /* ENUM */:
            (_e = visitor.enum) === null || _e === void 0 ? void 0 : _e.call(visitor, node);
            break;
        case 5 /* ELEMENTS */:
            (_f = visitor.elements) === null || _f === void 0 ? void 0 : _f.call(visitor, node, function () { return visitJtdNode(node.elementNode, visitor); });
            break;
        case 6 /* VALUES */:
            (_g = visitor.values) === null || _g === void 0 ? void 0 : _g.call(visitor, node, function () { return visitJtdNode(node.valueNode, visitor); });
            break;
        case 7 /* OBJECT */:
            (_h = visitor.object) === null || _h === void 0 ? void 0 : _h.call(visitor, node, function () { return visitJtdObjectNodeProperties(node, visitor); });
            break;
        case 8 /* UNION */:
            (_j = visitor.union) === null || _j === void 0 ? void 0 : _j.call(visitor, node, function () {
                var _a;
                var _loop_1 = function (mappingKey, mappingNode) {
                    (_a = visitor.mapping) === null || _a === void 0 ? void 0 : _a.call(visitor, mappingKey, mappingNode, node, function () { return visitJtdObjectNodeProperties(mappingNode, visitor); });
                };
                for (var _i = 0, _b = Object.entries(node.mapping); _i < _b.length; _i++) {
                    var _c = _b[_i], mappingKey = _c[0], mappingNode = _c[1];
                    _loop_1(mappingKey, mappingNode);
                }
            });
            break;
    }
}
function visitJtdObjectNodeProperties(node, visitor) {
    var _a, _b;
    var _loop_2 = function (propKey, propNode) {
        (_a = visitor.property) === null || _a === void 0 ? void 0 : _a.call(visitor, propKey, propNode, node, function () { return visitJtdNode(propNode, visitor); });
    };
    for (var _i = 0, _c = Object.entries(node.properties); _i < _c.length; _i++) {
        var _d = _c[_i], propKey = _d[0], propNode = _d[1];
        _loop_2(propKey, propNode);
    }
    var _loop_3 = function (propKey, propNode) {
        (_b = visitor.optionalProperty) === null || _b === void 0 ? void 0 : _b.call(visitor, propKey, propNode, node, function () { return visitJtdNode(propNode, visitor); });
    };
    for (var _e = 0, _f = Object.entries(node.optionalProperties); _e < _f.length; _e++) {
        var _g = _f[_e], propKey = _g[0], propNode = _g[1];
        _loop_3(propKey, propNode);
    }
}

var _a;
/**
 * Compiles provided JTD definitions as a TypeScript source.
 *
 * @param definitions Map from ref to JTD node.
 * @param options The compilation options.
 *
 * @returns The TypeScript source code with type, interface and enum definitions.
 */
function compileTsTypes(definitions, options) {
    var opts = __assign(__assign({}, tsTypesCompilerOptions), options);
    return Object.keys(definitions).reduce(function (src, ref) { return src + compileTsTypeStatement(ref, definitions, opts); }, '');
}
/**
 * Returns the TypeScript type statement that describes the node.
 *
 * @param ref The ref of the node in `definitions`.
 * @param definitions Known definitions that are used for ref resolution.
 * @param options Compiler options.
 */
function compileTsTypeStatement(ref, definitions, options) {
    var renameType = options.renameType, renamePropertyKey = options.renamePropertyKey, renameEnumKey = options.renameEnumKey, rewriteEnumValue = options.rewriteEnumValue, renameUnionEnum = options.renameUnionEnum, renameUnionEnumKey = options.renameUnionEnumKey, renameDiscriminatorKey = options.renameDiscriminatorKey, rewriteMappingKey = options.rewriteMappingKey, renameMappingInterface = options.renameMappingInterface, getDocComment = options.getDocComment;
    var src = '';
    var compileTypeStatement = function (node) {
        src += codegen.compileDocComment(getDocComment(node))
            + ("export type " + renameType(ref, node) + "=" + compileTsTypeExpression(node, definitions, options) + ";");
    };
    visitJtdNode(definitions[ref], {
        any: compileTypeStatement,
        ref: compileTypeStatement,
        nullable: compileTypeStatement,
        type: compileTypeStatement,
        elements: compileTypeStatement,
        values: compileTypeStatement,
        enum: function (node) {
            var name = renameType(ref, node);
            src += codegen.compileDocComment(getDocComment(node))
                + ("enum " + name + "{");
            for (var _i = 0, _a = node.values; _i < _a.length; _i++) {
                var value = _a[_i];
                src += renameEnumKey(value, node)
                    + '='
                    + JSON.stringify(rewriteEnumValue(value, node))
                    + ',';
            }
            src += '}'
                // Support of enum name mangling
                + ("export{" + name + "};");
        },
        object: function (node, next) {
            src += codegen.compileDocComment(getDocComment(node))
                + ("export interface " + renameType(ref, node) + "{");
            next();
            src += '}';
        },
        property: function (propKey, propNode, objectNode) {
            src += codegen.compileDocComment(getDocComment(propNode))
                + codegen.compilePropertyName(renamePropertyKey(propKey, propNode, objectNode))
                + ':'
                + compileTsTypeExpression(propNode, definitions, options)
                + ';';
        },
        optionalProperty: function (propKey, propNode, objectNode) {
            src += codegen.compileDocComment(getDocComment(propNode))
                + codegen.compilePropertyName(renamePropertyKey(propKey, propNode, objectNode))
                + '?:'
                + compileTsTypeExpression(propNode, definitions, options)
                + ';';
        },
        union: function (node, next) {
            var unionName = renameType(ref, node);
            var mappingEntries = Object.entries(node.mapping);
            if (mappingEntries.length === 0) {
                src += "export type " + unionName + "=never;";
                return;
            }
            var enumName = renameUnionEnum(ref, node);
            src += "enum " + enumName + "{"
                + mappingEntries.reduce(function (src, _a) {
                    var mappingKey = _a[0], mappingNode = _a[1];
                    return src
                        + renameUnionEnumKey(mappingKey, mappingNode, ref, node)
                        + '='
                        + JSON.stringify(rewriteMappingKey(mappingKey, mappingNode, ref, node))
                        + ',';
                }, '')
                + '}'
                // Support of enum name mangling
                + ("export{" + enumName + "};")
                + codegen.compileDocComment(getDocComment(node))
                + ("export type " + unionName + "=")
                + mappingEntries.reduce(function (src, _a) {
                    var mappingKey = _a[0], mappingNode = _a[1];
                    return src
                        + '|' + renameMappingInterface(mappingKey, mappingNode, ref, node);
                }, '')
                + ';';
            next();
        },
        mapping: function (mappingKey, mappingNode, unionNode, next) {
            src += codegen.compileDocComment(getDocComment(mappingNode))
                + ("export interface " + renameMappingInterface(mappingKey, mappingNode, ref, unionNode) + "{")
                + codegen.compilePropertyName(renameDiscriminatorKey(unionNode))
                + ':'
                + renameUnionEnum(ref, unionNode) + '.' + renameUnionEnumKey(mappingKey, mappingNode, ref, unionNode)
                + ';';
            next();
            src += '}';
        },
    });
    return src;
}
/**
 * Returns the TypeScript type expression that describes the node.
 *
 * @param node The JTD node for which TypeScript expression must be compiled.
 * @param definitions Known definitions that are used for ref resolution.
 * @param options Compiler options.
 */
function compileTsTypeExpression(node, definitions, options) {
    var resolveExternalRef = options.resolveExternalRef, renameType = options.renameType, rewritePrimitiveType = options.rewritePrimitiveType, rewriteEnumValue = options.rewriteEnumValue, rewriteMappingKey = options.rewriteMappingKey, renameDiscriminatorKey = options.renameDiscriminatorKey, getDocComment = options.getDocComment;
    var src = '';
    visitJtdNode(node, {
        any: function () {
            src += 'any';
        },
        ref: function (node) {
            src += definitions[node.ref] ? renameType(node.ref, definitions[node.ref]) : resolveExternalRef(node);
        },
        nullable: function (node, next) {
            next();
            src += '|null';
        },
        type: function (node) {
            src += rewritePrimitiveType(node);
        },
        enum: function (node) {
            src += node.values.reduce(function (src, value) { return src + '|' + JSON.stringify(rewriteEnumValue(value, node)); }, '');
        },
        elements: function (node, next) {
            src += 'Array<';
            next();
            src += '>';
        },
        values: function (node, next) {
            src += 'Record<string,';
            next();
            src += '>';
        },
        object: function (node, next) {
            src += '{';
            next();
            src += '}';
        },
        property: function (propKey, propNode, objectNode, next) {
            src += codegen.compileDocComment(getDocComment(propNode))
                + codegen.compilePropertyName(propKey)
                + ':';
            next();
            src += ';';
        },
        optionalProperty: function (propKey, propNode, objectNode, next) {
            src += codegen.compileDocComment(getDocComment(propNode))
                + codegen.compilePropertyName(propKey)
                + '?:';
            next();
            src += ';';
        },
        union: function (node, next) {
            if (Object.keys(node.mapping).length === 0) {
                src += 'never';
            }
            else {
                next();
            }
        },
        mapping: function (mappingKey, mappingNode, unionNode, next) {
            src += '|{'
                + renameDiscriminatorKey(unionNode)
                + ':'
                + JSON.stringify(rewriteMappingKey(mappingKey, mappingNode, undefined, unionNode))
                + ';';
            next();
            src += '}';
        },
    });
    return src;
}
/**
 * Global default options used by {@link compileTsTypes}.
 */
var tsTypesCompilerOptions = {
    resolveExternalRef: function (node) { return die('Unresolved reference: ' + node.ref); },
    renameType: function (ref) { return changeCaseAll.pascalCase(ref); },
    renamePropertyKey: function (propKey) { return propKey; },
    rewritePrimitiveType: function (node) { return jtdTsPrimitiveTypeMap[node.type] || die('Unknown type: ' + node.type); },
    renameEnumKey: function (ref) { return changeCaseAll.constantCase(ref); },
    rewriteEnumValue: function (value) { return value; },
    renameUnionEnum: function (ref, node) { return changeCaseAll.pascalCase(ref) + changeCaseAll.pascalCase(node.discriminator); },
    renameUnionEnumKey: function (ref) { return changeCaseAll.constantCase(ref); },
    renameDiscriminatorKey: function (node) { return node.discriminator; },
    rewriteMappingKey: function (mappingKey) { return mappingKey; },
    renameMappingInterface: function (mappingKey, mappingNode, unionRef) { return changeCaseAll.pascalCase(unionRef) + changeCaseAll.pascalCase(mappingKey); },
    getDocComment: function (node) { var _a; return (_a = node.jtd.metadata) === null || _a === void 0 ? void 0 : _a.comment; },
};
/**
 * Mapping from the JTD standard data types to TypeScript primitive types.
 */
var jtdTsPrimitiveTypeMap = (_a = {},
    _a["boolean" /* BOOLEAN */] = 'boolean',
    _a["string" /* STRING */] = 'string',
    _a["timestamp" /* TIMESTAMP */] = 'number',
    _a["float32" /* FLOAT32 */] = 'number',
    _a["float64" /* FLOAT64 */] = 'number',
    _a["int8" /* INT8 */] = 'number',
    _a["uint8" /* UINT8 */] = 'number',
    _a["int16" /* INT16 */] = 'number',
    _a["uint16" /* UINT16 */] = 'number',
    _a["int32" /* INT32 */] = 'number',
    _a["uint32" /* UINT32 */] = 'number',
    _a);

/**
 * Compiles validators and type guards from the definitions.
 */
function compileValidators(definitions, options) {
    var opt = __assign(__assign({}, validatorCompilerOptions), options);
    var typeGuardsRendered = opt.typeGuardsRendered, dialect = opt.dialect;
    var src = '';
    var _loop_1 = function (ref, node) {
        src += codegen.compileJsSource(dialect.validator(ref, node, function (ctx) { return compileValidatorBody(node, ctx, dialect); }));
        if (typeGuardsRendered) {
            src += codegen.compileJsSource(dialect.typeGuard(ref, node));
        }
    };
    for (var _i = 0, _a = Object.entries(definitions); _i < _a.length; _i++) {
        var _b = _a[_i], ref = _b[0], node = _b[1];
        _loop_1(ref, node);
    }
    return src;
}
function compileValidatorBody(node, ctx, dialect) {
    var frags = [];
    var createNext = function (next) { return function (nextCtx) {
        var prevFrags = frags;
        var prevCtx = ctx;
        frags = [];
        ctx = nextCtx;
        next();
        var nextFrag = codegen.template(frags);
        frags = prevFrags;
        ctx = prevCtx;
        return nextFrag;
    }; };
    visitJtdNode(node, {
        ref: function (node) {
            frags.push(dialect.ref(node, ctx));
        },
        nullable: function (node, next) {
            frags.push(dialect.nullable(node, ctx, createNext(next)));
        },
        type: function (node) {
            frags.push(dialect.type(node, ctx));
        },
        enum: function (node) {
            frags.push(dialect.enum(node, ctx));
        },
        elements: function (node, next) {
            frags.push(dialect.elements(node, ctx, createNext(next)));
        },
        values: function (node, next) {
            frags.push(dialect.values(node, ctx, createNext(next)));
        },
        object: function (node, next) {
            frags.push(dialect.object(node, ctx, createNext(next)));
        },
        property: function (propKey, propNode, objectNode, next) {
            frags.push(dialect.property(propKey, propNode, objectNode, ctx, createNext(next)));
        },
        optionalProperty: function (propKey, propNode, objectNode, next) {
            frags.push(dialect.optionalProperty(propKey, propNode, objectNode, ctx, createNext(next)));
        },
        union: function (node, next) {
            frags.push(dialect.union(node, ctx, createNext(next)));
        },
        mapping: function (mappingKey, mappingNode, unionNode, next) {
            frags.push(dialect.mapping(mappingKey, mappingNode, unionNode, ctx, createNext(next)));
        },
    });
    return codegen.template(frags);
}
var validatorCompilerOptions = {
    typeGuardsRendered: false,
    dialect: jtdDialect.createJtdDialect(),
};

/**
 * Compiles map of JTD definitions to a map of corresponding sources.
 *
 * @param jtdModules The map from module URI to JTD definitions map.
 * @param options Compiler options.
 */
function compileTsModules(jtdModules, options) {
    var opts = __assign(__assign(__assign(__assign({}, validatorCompilerOptions), tsTypesCompilerOptions), jtdDialect.jtdDialectOptions), options);
    var validatorsRendered = opts.validatorsRendered, _a = opts.dialectFactory, dialectFactory = _a === void 0 ? jtdDialect.createJtdDialect : _a;
    var dialect = dialectFactory(opts);
    var tsModules = createMap();
    var modules = Object.entries(jtdModules).map(function (_a) {
        var uri = _a[0], jtdDefinitions = _a[1];
        var definitions = parseJtdDefinitions(jtdDefinitions);
        var exports = extractExports(definitions, opts);
        return { uri: uri, definitions: definitions, exports: exports };
    });
    var _loop_1 = function (module_1) {
        var imports = createMap();
        opts.resolveExternalRef = function (node) {
            var _a;
            for (var _i = 0, modules_2 = modules; _i < modules_2.length; _i++) {
                var module_2 = modules_2[_i];
                var exportedRef = module_2.exports[node.ref];
                if (exportedRef) {
                    var importedNames = imports[_a = module_2.uri] || (imports[_a] = new Set());
                    importedNames.add(exportedRef.typeName);
                    if (validatorsRendered) {
                        importedNames.add(exportedRef.validatorName);
                    }
                    return exportedRef.typeName;
                }
            }
            die('Unresolved reference: ' + node.ref);
        };
        var src = compileTsTypes(module_1.definitions, opts);
        for (var _b = 0, _c = Object.entries(imports); _b < _c.length; _b++) {
            var _d = _c[_b], uri = _d[0], importedNames = _d[1];
            src = "import{" + Array.from(importedNames).sort().join(',') + "}from" + JSON.stringify(uri) + ";" + src;
        }
        if (validatorsRendered && dialect != null) {
            src = codegen.compileJsSource(dialect.import()) + src;
            src += compileValidators(module_1.definitions, opts);
        }
        tsModules[module_1.uri] = src;
    };
    for (var _i = 0, modules_1 = modules; _i < modules_1.length; _i++) {
        var module_1 = modules_1[_i];
        _loop_1(module_1);
    }
    return tsModules;
}
function extractExports(definitions, options) {
    var exports = createMap();
    for (var _i = 0, _a = Object.entries(definitions); _i < _a.length; _i++) {
        var _b = _a[_i], ref = _b[0], node = _b[1];
        exports[ref] = {
            typeName: options.renameType(ref, node),
            validatorName: options.renameValidator(ref, node),
        };
    }
    return exports;
}

exports.compileTsModules = compileTsModules;
exports.compileTsTypes = compileTsTypes;
exports.compileValidators = compileValidators;
exports.jtdTsPrimitiveTypeMap = jtdTsPrimitiveTypeMap;
exports.parseJtd = parseJtd;
exports.parseJtdDefinitions = parseJtdDefinitions;
exports.parseJtdRoot = parseJtdRoot;
exports.tsTypesCompilerOptions = tsTypesCompilerOptions;
exports.validatorCompilerOptions = validatorCompilerOptions;
exports.visitJtdNode = visitJtdNode;
