import { __assign } from "tslib";
import { createMap, die } from './misc';
/**
 * Converts JTD and its dependencies to a map of nodes where key is `ref` and value is a parsed node.
 *
 * @param ref The ref of the root JTD.
 * @param jtdRoot The JTD to parse.
 *
 * @returns The map from ref to a parsed node.
 */
export function parseJtdRoot(ref, jtdRoot) {
    var nodes = jtdRoot.definitions ? parseJtdDefinitions(jtdRoot.definitions) : createMap();
    nodes[ref] = parseJtd(jtdRoot);
    return nodes;
}
/**
 * Converts JTD dependencies to a map of nodes where key is `ref` and value is a parsed node.
 *
 * @param jtdDefinitions The dictionary of ref-JTD pairs.
 *
 * @returns The map from ref to a parsed node.
 */
export function parseJtdDefinitions(jtdDefinitions) {
    var nodes = createMap();
    for (var _i = 0, _a = Object.entries(jtdDefinitions); _i < _a.length; _i++) {
        var _b = _a[_i], ref = _b[0], jtd = _b[1];
        nodes[ref] = parseJtd(jtd);
    }
    return nodes;
}
/**
 * Converts JTD to a corresponding node.
 *
 * @param jtd The JTD to parse.
 *
 * @returns A parsed node.
 *
 * @see https://tools.ietf.org/html/rfc8927 RFC8927
 * @see https://jsontypedef.com/docs/jtd-in-5-minutes JTD in 5 minutes
 */
export function parseJtd(jtd) {
    var jtdNullable = jtd.nullable, jtdType = jtd.type, jtdRef = jtd.ref, jtdEnum = jtd.enum, jtdElements = jtd.elements, jtdValues = jtd.values, jtdProperties = jtd.properties, jtdOptionalProperties = jtd.optionalProperties, jtdDiscriminator = jtd.discriminator, jtdMapping = jtd.mapping;
    if (jtdNullable) {
        var node = {
            nodeType: 2 /* NULLABLE */,
            valueNode: parseJtd(__assign(__assign({}, jtd), { nullable: undefined })),
            parentNode: null,
            jtd: jtd,
        };
        node.valueNode.parentNode = node;
        return node;
    }
    if (jtdType) {
        return {
            nodeType: 3 /* TYPE */,
            type: jtdType,
            parentNode: null,
            jtd: jtd,
        };
    }
    if (jtdRef) {
        return {
            nodeType: 1 /* REF */,
            ref: jtdRef,
            parentNode: null,
            jtd: jtd,
        };
    }
    if (jtdEnum) {
        return {
            nodeType: 4 /* ENUM */,
            values: jtdEnum,
            parentNode: null,
            jtd: jtd,
        };
    }
    if (jtdElements) {
        var node = {
            nodeType: 5 /* ELEMENTS */,
            elementNode: parseJtd(jtdElements),
            parentNode: null,
            jtd: jtd,
        };
        node.elementNode.parentNode = node;
        return node;
    }
    if (jtdValues) {
        var node = {
            nodeType: 6 /* VALUES */,
            valueNode: parseJtd(jtdValues),
            parentNode: null,
            jtd: jtd,
        };
        node.valueNode.parentNode = node;
        return node;
    }
    if (jtdProperties || jtdOptionalProperties) {
        var objectNode = {
            nodeType: 7 /* OBJECT */,
            properties: createMap(),
            optionalProperties: createMap(),
            parentNode: null,
            jtd: jtd,
        };
        if (jtdProperties) {
            for (var _i = 0, _a = Object.entries(jtdProperties); _i < _a.length; _i++) {
                var _b = _a[_i], propKey = _b[0], propJtd = _b[1];
                var propNode = objectNode.properties[propKey] = parseJtd(propJtd);
                propNode.parentNode = objectNode;
            }
        }
        if (jtdOptionalProperties) {
            for (var _c = 0, _d = Object.entries(jtdOptionalProperties); _c < _d.length; _c++) {
                var _e = _d[_c], propKey = _e[0], propJtd = _e[1];
                if (propKey in objectNode.properties) {
                    die('Duplicated property: ' + propKey);
                }
                var propNode = objectNode.optionalProperties[propKey] = parseJtd(propJtd);
                propNode.parentNode = objectNode;
            }
        }
        return objectNode;
    }
    if (jtdDiscriminator || jtdMapping) {
        if (!jtdDiscriminator || !jtdMapping) {
            die('Malformed discriminated union');
        }
        var unionNode = {
            nodeType: 8 /* UNION */,
            discriminator: jtdDiscriminator,
            mapping: createMap(),
            parentNode: null,
            jtd: jtd,
        };
        for (var _f = 0, _g = Object.entries(jtdMapping); _f < _g.length; _f++) {
            var _h = _g[_f], mappingKey = _h[0], mappingJtd = _h[1];
            var objectNode = parseJtd(mappingJtd);
            if (objectNode.nodeType !== 7 /* OBJECT */) {
                die('Mappings must be object definitions: ' + mappingKey);
            }
            unionNode.mapping[mappingKey] = objectNode;
            objectNode.parentNode = unionNode;
        }
        return unionNode;
    }
    return {
        nodeType: 0 /* ANY */,
        parentNode: null,
        jtd: jtd,
    };
}
