import { IJtdEnumNode, IJtdObjectNode, IJtdRefNode, IJtdTypeNode, IJtdUnionNode, JtdNode } from '@jtdc/types';
export interface ITsTypesCompilerOptions<M> {
    /**
     * Returns a TypeScript type name referenced by `node`.
     */
    resolveExternalRef?: (node: IJtdRefNode<M>) => string;
    /**
     * Returns a TypeScript type name described by `node`.
     *
     * @param ref The ref of the renamed type.
     * @param node The node that describes the renamed type.
     */
    renameType?: (ref: string, node: JtdNode<M>) => string;
    /**
     * Returns a TypeScript type name of a JTD primitive type. By default, a {@link jtdTsPrimitiveTypeMap} is used to
     * resolve the name.
     *
     * @example
     * (node) => node.type === 'int64' ? 'bigint' : 'string'
     */
    rewritePrimitiveType?: (node: IJtdTypeNode<M>) => string;
    /**
     * Returns the new name of an object property.
     */
    renamePropertyKey?: (propKey: string, propNode: JtdNode<M>, objectNode: IJtdObjectNode<M>) => string;
    /**
     * Returns the name of the enum key.
     */
    renameEnumKey?: (value: string, node: IJtdEnumNode<M>) => string;
    /**
     * Returns the contents of the enum value.
     */
    rewriteEnumValue?: (value: string, node: IJtdEnumNode<M>) => string | number | undefined;
    /**
     * Returns the name of the enum that holds mapping keys for the discriminated union.
     */
    renameUnionEnum?: (ref: string, node: IJtdUnionNode<M>) => string;
    /**
     * Returns the contents of the value from the enum that holds discriminator values.
     */
    renameUnionEnumKey?: (mappingKey: string, mappingNode: IJtdObjectNode<M>, unionRef: string, unionNode: IJtdUnionNode<M>) => string;
    /**
     * Returns the name of the union discriminator property.
     */
    renameDiscriminatorKey?: (node: IJtdUnionNode<M>) => string;
    /**
     * Returns the value that would be used as a value of the discriminator property.
     */
    rewriteMappingKey?: (mappingKey: string, mappingNode: IJtdObjectNode<M>, unionRef: string | undefined, unionNode: IJtdUnionNode<M>) => string | number | undefined;
    /**
     * Returns the name of the interface that is the part of the discriminated union.
     */
    renameMappingInterface?: (mappingKey: string, mappingNode: IJtdObjectNode<M>, unionRef: string, unionNode: IJtdUnionNode<M>) => string;
    /**
     * Returns the doc comment string that is associated with the node.
     *
     * @default node.metadata?.comment
     */
    getDocComment?: (node: JtdNode<M>) => string | null | undefined;
}
/**
 * Compiles provided JTD definitions as a TypeScript source.
 *
 * @param definitions Map from ref to JTD node.
 * @param options The compilation options.
 *
 * @returns The TypeScript source code with type, interface and enum definitions.
 */
export declare function compileTsTypes<M>(definitions: Record<string, JtdNode<M>>, options?: ITsTypesCompilerOptions<M>): string;
/**
 * Global default options used by {@link compileTsTypes}.
 */
export declare const tsTypesCompilerOptions: Required<ITsTypesCompilerOptions<any>>;
/**
 * Mapping from the JTD standard data types to TypeScript primitive types.
 */
export declare const jtdTsPrimitiveTypeMap: Record<string, string>;
