var _a;
import { __assign } from "tslib";
import { visitJtdNode } from './jtd-visitor';
import { compileDocComment, compilePropertyName } from '@smikhalevski/codegen';
import { constantCase, pascalCase } from 'change-case-all';
import { die } from './misc';
/**
 * Compiles provided JTD definitions as a TypeScript source.
 *
 * @param definitions Map from ref to JTD node.
 * @param options The compilation options.
 *
 * @returns The TypeScript source code with type, interface and enum definitions.
 */
export function compileTsTypes(definitions, options) {
    var opts = __assign(__assign({}, tsTypesCompilerOptions), options);
    return Object.keys(definitions).reduce(function (src, ref) { return src + compileTsTypeStatement(ref, definitions, opts); }, '');
}
/**
 * Returns the TypeScript type statement that describes the node.
 *
 * @param ref The ref of the node in `definitions`.
 * @param definitions Known definitions that are used for ref resolution.
 * @param options Compiler options.
 */
function compileTsTypeStatement(ref, definitions, options) {
    var renameType = options.renameType, renamePropertyKey = options.renamePropertyKey, renameEnumKey = options.renameEnumKey, rewriteEnumValue = options.rewriteEnumValue, renameUnionEnum = options.renameUnionEnum, renameUnionEnumKey = options.renameUnionEnumKey, renameDiscriminatorKey = options.renameDiscriminatorKey, rewriteMappingKey = options.rewriteMappingKey, renameMappingInterface = options.renameMappingInterface, getDocComment = options.getDocComment;
    var src = '';
    var compileTypeStatement = function (node) {
        src += compileDocComment(getDocComment(node))
            + ("export type " + renameType(ref, node) + "=" + compileTsTypeExpression(node, definitions, options) + ";");
    };
    visitJtdNode(definitions[ref], {
        any: compileTypeStatement,
        ref: compileTypeStatement,
        nullable: compileTypeStatement,
        type: compileTypeStatement,
        elements: compileTypeStatement,
        values: compileTypeStatement,
        enum: function (node) {
            var name = renameType(ref, node);
            src += compileDocComment(getDocComment(node))
                + ("enum " + name + "{");
            for (var _i = 0, _a = node.values; _i < _a.length; _i++) {
                var value = _a[_i];
                src += renameEnumKey(value, node)
                    + '='
                    + JSON.stringify(rewriteEnumValue(value, node))
                    + ',';
            }
            src += '}'
                // Support of enum name mangling
                + ("export{" + name + "};");
        },
        object: function (node, next) {
            src += compileDocComment(getDocComment(node))
                + ("export interface " + renameType(ref, node) + "{");
            next();
            src += '}';
        },
        property: function (propKey, propNode, objectNode) {
            src += compileDocComment(getDocComment(propNode))
                + compilePropertyName(renamePropertyKey(propKey, propNode, objectNode))
                + ':'
                + compileTsTypeExpression(propNode, definitions, options)
                + ';';
        },
        optionalProperty: function (propKey, propNode, objectNode) {
            src += compileDocComment(getDocComment(propNode))
                + compilePropertyName(renamePropertyKey(propKey, propNode, objectNode))
                + '?:'
                + compileTsTypeExpression(propNode, definitions, options)
                + ';';
        },
        union: function (node, next) {
            var unionName = renameType(ref, node);
            var mappingEntries = Object.entries(node.mapping);
            if (mappingEntries.length === 0) {
                src += "export type " + unionName + "=never;";
                return;
            }
            var enumName = renameUnionEnum(ref, node);
            src += "enum " + enumName + "{"
                + mappingEntries.reduce(function (src, _a) {
                    var mappingKey = _a[0], mappingNode = _a[1];
                    return src
                        + renameUnionEnumKey(mappingKey, mappingNode, ref, node)
                        + '='
                        + JSON.stringify(rewriteMappingKey(mappingKey, mappingNode, ref, node))
                        + ',';
                }, '')
                + '}'
                // Support of enum name mangling
                + ("export{" + enumName + "};")
                + compileDocComment(getDocComment(node))
                + ("export type " + unionName + "=")
                + mappingEntries.reduce(function (src, _a) {
                    var mappingKey = _a[0], mappingNode = _a[1];
                    return src
                        + '|' + renameMappingInterface(mappingKey, mappingNode, ref, node);
                }, '')
                + ';';
            next();
        },
        mapping: function (mappingKey, mappingNode, unionNode, next) {
            src += compileDocComment(getDocComment(mappingNode))
                + ("export interface " + renameMappingInterface(mappingKey, mappingNode, ref, unionNode) + "{")
                + compilePropertyName(renameDiscriminatorKey(unionNode))
                + ':'
                + renameUnionEnum(ref, unionNode) + '.' + renameUnionEnumKey(mappingKey, mappingNode, ref, unionNode)
                + ';';
            next();
            src += '}';
        },
    });
    return src;
}
/**
 * Returns the TypeScript type expression that describes the node.
 *
 * @param node The JTD node for which TypeScript expression must be compiled.
 * @param definitions Known definitions that are used for ref resolution.
 * @param options Compiler options.
 */
function compileTsTypeExpression(node, definitions, options) {
    var resolveExternalRef = options.resolveExternalRef, renameType = options.renameType, rewritePrimitiveType = options.rewritePrimitiveType, rewriteEnumValue = options.rewriteEnumValue, rewriteMappingKey = options.rewriteMappingKey, renameDiscriminatorKey = options.renameDiscriminatorKey, getDocComment = options.getDocComment;
    var src = '';
    visitJtdNode(node, {
        any: function () {
            src += 'any';
        },
        ref: function (node) {
            src += definitions[node.ref] ? renameType(node.ref, definitions[node.ref]) : resolveExternalRef(node);
        },
        nullable: function (node, next) {
            next();
            src += '|null';
        },
        type: function (node) {
            src += rewritePrimitiveType(node);
        },
        enum: function (node) {
            src += node.values.reduce(function (src, value) { return src + '|' + JSON.stringify(rewriteEnumValue(value, node)); }, '');
        },
        elements: function (node, next) {
            src += 'Array<';
            next();
            src += '>';
        },
        values: function (node, next) {
            src += 'Record<string,';
            next();
            src += '>';
        },
        object: function (node, next) {
            src += '{';
            next();
            src += '}';
        },
        property: function (propKey, propNode, objectNode, next) {
            src += compileDocComment(getDocComment(propNode))
                + compilePropertyName(propKey)
                + ':';
            next();
            src += ';';
        },
        optionalProperty: function (propKey, propNode, objectNode, next) {
            src += compileDocComment(getDocComment(propNode))
                + compilePropertyName(propKey)
                + '?:';
            next();
            src += ';';
        },
        union: function (node, next) {
            if (Object.keys(node.mapping).length === 0) {
                src += 'never';
            }
            else {
                next();
            }
        },
        mapping: function (mappingKey, mappingNode, unionNode, next) {
            src += '|{'
                + renameDiscriminatorKey(unionNode)
                + ':'
                + JSON.stringify(rewriteMappingKey(mappingKey, mappingNode, undefined, unionNode))
                + ';';
            next();
            src += '}';
        },
    });
    return src;
}
/**
 * Global default options used by {@link compileTsTypes}.
 */
export var tsTypesCompilerOptions = {
    resolveExternalRef: function (node) { return die('Unresolved reference: ' + node.ref); },
    renameType: function (ref) { return pascalCase(ref); },
    renamePropertyKey: function (propKey) { return propKey; },
    rewritePrimitiveType: function (node) { return jtdTsPrimitiveTypeMap[node.type] || die('Unknown type: ' + node.type); },
    renameEnumKey: function (ref) { return constantCase(ref); },
    rewriteEnumValue: function (value) { return value; },
    renameUnionEnum: function (ref, node) { return pascalCase(ref) + pascalCase(node.discriminator); },
    renameUnionEnumKey: function (ref) { return constantCase(ref); },
    renameDiscriminatorKey: function (node) { return node.discriminator; },
    rewriteMappingKey: function (mappingKey) { return mappingKey; },
    renameMappingInterface: function (mappingKey, mappingNode, unionRef) { return pascalCase(unionRef) + pascalCase(mappingKey); },
    getDocComment: function (node) { var _a; return (_a = node.jtd.metadata) === null || _a === void 0 ? void 0 : _a.comment; },
};
/**
 * Mapping from the JTD standard data types to TypeScript primitive types.
 */
export var jtdTsPrimitiveTypeMap = (_a = {},
    _a["boolean" /* BOOLEAN */] = 'boolean',
    _a["string" /* STRING */] = 'string',
    _a["timestamp" /* TIMESTAMP */] = 'number',
    _a["float32" /* FLOAT32 */] = 'number',
    _a["float64" /* FLOAT64 */] = 'number',
    _a["int8" /* INT8 */] = 'number',
    _a["uint8" /* UINT8 */] = 'number',
    _a["int16" /* INT16 */] = 'number',
    _a["uint16" /* UINT16 */] = 'number',
    _a["int32" /* INT32 */] = 'number',
    _a["uint32" /* UINT32 */] = 'number',
    _a);
